// ----------------------------------------------------------------------------
// (C) 2002 
//     carrara engineering
//     Switzerland
//     http://www.carrara.ch
//     mailto:info@carrara.ch
// ----------------------------------------------------------------------------
// Permission to use, copy, modify, distribute and sell this software for any 
//     purpose is hereby granted without fee, provided that the above copyright 
//     notice appear in all copies and that both that copyright notice and this 
//     permission notice appear in supporting documentation.
// The author make no representations about the suitability 
//     of this software for any purpose. It is provided "as is" 
//     without express or implied warranty.
// ----------------------------------------------------------------------------

#include <iostream.h>
#include <stdlib.h>

// ----------------------------------------------------------------------------

namespace LeakHunter1
{

class HeapControl
{
  public:
    void* operator new(size_t i_nSize)
    {
      void* paObject = malloc(i_nSize);
      if (paObject)
      {
        ++sm_nCounter;
      }

      return paObject;
    }

    void operator delete(void* i_paObject)
    {
      if (i_paObject)
      {
        free(i_paObject);
        --sm_nCounter;
      }
    }

    static int sm_nCounter;

};

int HeapControl::sm_nCounter=0;

class A : public HeapControl
{
  int m_nInt;
};

class B : public A
{
  float m_nFloat;
};

class C : public HeapControl
{
  double m_nDouble;
};

void test()
{
  A* paA = new A();
  A* paB = new B();
  C* paC = new C();

  delete paC;

  cout << "LeakHunter: " << HeapControl::sm_nCounter << " leaks detected!" << endl;

}

}

// ----------------------------------------------------------------------------

namespace LeakHunter2
{
class HeapMgr
{
  public:

    static HeapMgr& instance()
    {
      if (!ms_paInstance)
      {
        ms_paInstance = new HeapMgr();
      }
      return *ms_paInstance;
    }

    void* alloc(size_t i_nSize)
    {
      void* paObject = malloc(i_nSize);
      if (paObject)
      {
        ++m_nCounter;
      }

      return paObject;
    }

    void free(void* i_paObject)
    {
      if (i_paObject)
      {
        ::free(i_paObject);
        --m_nCounter;
      }
    }

    void printStatus()
    {
      cout << "LeakHunter: " << m_nCounter << " leaks detected!" << endl;
    }

  private:
    HeapMgr()
    : m_nCounter (0)
    {}

    int m_nCounter;

    static HeapMgr* ms_paInstance;
};

HeapMgr* HeapMgr::ms_paInstance=0;

class HeapControl
{
  public:
    void* operator new(size_t i_nSize)
    {
      return HeapMgr::instance().alloc(i_nSize);
    }

    void operator delete(void* i_paObject)
    {
      HeapMgr::instance().free(i_paObject);
    }

};

class A : public HeapControl
{
  int m_nInt;
};

class B : public A
{
  float m_nFloat;
};

class C : public HeapControl
{
  double m_nDouble;
};

void test()
{
  A* paA = new A();
  A* paB = new B();
  C* paC = new C();

  delete paC;

  HeapMgr::instance().printStatus();
}

}

// ----------------------------------------------------------------------------

namespace LeakHunter3
{

class HeapMgr
{
  public:

    static HeapMgr& instance()
    {
      if (!ms_paInstance)
      {
        ms_paInstance = new HeapMgr();
      }
      return *ms_paInstance;
    }

    void* alloc(size_t i_nSize, const char* i_pcFile, int i_nLine)
    {
      void* paObject = malloc(i_nSize);
      if (paObject)
      {
        cout << "LeakHunter: " << i_nSize << " bytes allocated @ " << i_pcFile << ":" << i_nLine << endl;
        ++m_nCounter;
      }

      return paObject;
    }

    void free(void* i_paObject)
    {
      if (i_paObject)
      {
        ::free(i_paObject);
        --m_nCounter;
      }
    }

    void printStatus()
    {
      cout << "LeakHunter: " << m_nCounter << " leaks detected!" << endl;
    }

  private:
    HeapMgr()
    : m_nCounter (0)
    {}

    int m_nCounter;

    static HeapMgr* ms_paInstance;
};

HeapMgr* HeapMgr::ms_paInstance=0;


class HeapControl
{
  public:
    void* operator new(size_t i_nSize, const char* i_pcFile, int i_nLine)
    {
      return HeapMgr::instance().alloc(i_nSize, i_pcFile, i_nLine);
    }

    void operator delete(void* i_paObject)
    {
      HeapMgr::instance().free(i_paObject);
    }

};

#define new new (__FILE__,__LINE__) 

class A : public HeapControl
{
  int m_nInt;
};

class B : public A
{
  float m_nFloat;
};

class C : public HeapControl
{
  double m_nDouble;
};

void test()
{
  A* paA = new A();
  A* paB = new B();
  C* paC = new C();

  delete paC;

  HeapMgr::instance().printStatus();
}

#undef new

}

// ----------------------------------------------------------------------------

namespace LeakHunter4
{

class HeapMgr
{
  private:

    class Info
    {
      public:
        const char* m_pcFile;
        int m_nLine;
    };

  public:

    static HeapMgr& instance()
    {
      if (!ms_paInstance)
      {
        ms_paInstance = new HeapMgr();
      }
      return *ms_paInstance;
    }

    void* alloc(size_t i_nSize, const char* i_pcFile, int i_nLine)
    {
      void* paObject = malloc(i_nSize+sizeof(Info));
      if (paObject)
      {
        ++m_nCounter;
        Info* paInfo = (Info*)paObject;
        paInfo->m_pcFile = i_pcFile;
        paInfo->m_nLine  = i_nLine;
        cout << "LeakHunter: " << i_nSize << " bytes allocated @ " << paInfo->m_pcFile 
             << ":" << paInfo->m_nLine << endl;
      }
      return (char*)paObject+sizeof(Info);
    }

    void free(void* i_paObject)
    {
      if (i_paObject)
      {
        Info* paInfo = (Info*)((char*)i_paObject-sizeof(Info));
        cout << "LeakHunter: " << "Object freed @ " << paInfo->m_pcFile 
             << ":" << paInfo->m_nLine << endl;
        ::free(paInfo);
        --m_nCounter;
      }
    }

    void printStatus()
    {
      cout << "LeakHunter: " << m_nCounter << " leaks detected!" << endl;
    }

  private:
    HeapMgr()
    : m_nCounter (0)
    {}

    int m_nCounter;

    static HeapMgr* ms_paInstance;
};

HeapMgr* HeapMgr::ms_paInstance=0;


class HeapControl
{
  public:
    void* operator new(size_t i_nSize, const char* i_pcFile, int i_nLine)
    {
      return HeapMgr::instance().alloc(i_nSize, i_pcFile, i_nLine);
    }

    void operator delete(void* i_paObject)
    {
      HeapMgr::instance().free(i_paObject);
    }

};

#define new new (__FILE__,__LINE__) 

class A : public HeapControl
{
  int m_nInt;
};

class B : public A
{
  float m_nFloat;
};

class C : public HeapControl
{
  double m_nDouble;
};

void test()
{
  A* paA = new A();
  A* paB = new B();
  C* paC = new C();

  delete paC;

  HeapMgr::instance().printStatus();
}

#undef new


}

// ----------------------------------------------------------------------------

namespace LeakHunter5
{

class HeapMgr
{
  private:

    class Info
    {
      public:
        const char* m_pcFile;
        int m_nLine;
        Info* m_paNext;
        Info* m_paPrev;
    };

  public:

    static HeapMgr& instance()
    {
      if (!ms_paInstance)
      {
        ms_paInstance = new HeapMgr();
      }
      return *ms_paInstance;
    }

    void* alloc(size_t i_nSize, const char* i_pcFile, int i_nLine)
    {
      void* paObject = malloc(i_nSize+sizeof(Info));
      if (paObject)
      {
        Info* paInfo = (Info*)paObject;
        paInfo->m_pcFile = i_pcFile;
        paInfo->m_nLine  = i_nLine;
        
        if (m_paFront)
        {
          paInfo->m_paPrev = m_paBack;
          paInfo->m_paNext = 0;
          m_paBack = paInfo;
          paInfo->m_paPrev->m_paNext = paInfo;
        }
        else
        {
          m_paFront = paInfo;
          m_paBack  = paInfo;
          paInfo->m_paNext = 0;
          paInfo->m_paPrev = 0;
        }

      }
      return (char*)paObject+sizeof(Info);
    }

    void free(void* i_paObject)
    {
      if (i_paObject)
      {
        Info* paInfo = (Info*)((char*)i_paObject-sizeof(Info));

        if (paInfo->m_paPrev) 
        {
          paInfo->m_paPrev->m_paNext=paInfo->m_paNext;
        }
        else
        {
          m_paFront = paInfo->m_paNext;
        }

        if (paInfo->m_paNext) 
        {
          paInfo->m_paNext->m_paPrev=paInfo->m_paPrev;
        }
        else
        {
          m_paBack = paInfo->m_paPrev;
        }

        paInfo->m_paNext = 0;
        paInfo->m_paPrev = 0;
        ::free(paInfo);
      }
    }

    void printStatus()
    {
      if (m_paFront)
      {
        int nCounter=0;
        cout << "Reporting: " << endl;
        Info* paInfo = m_paFront;
        while (paInfo)
        {
          cout << "LeakHunter: " << "Object lost @ " << paInfo->m_pcFile 
               << ":" << paInfo->m_nLine << endl;
          paInfo = paInfo->m_paNext;
          ++nCounter;
        }
        cout << "LeakHunter: " << nCounter << " leaks detected!" << endl;
      }
    }

  private:
    
    HeapMgr()
    : m_paFront  (0)
    , m_paBack   (0)
    {}

    Info* m_paFront;
    Info* m_paBack;

    static HeapMgr* ms_paInstance;
};

HeapMgr* HeapMgr::ms_paInstance=0;


class HeapControl
{
  public:
    
    void* operator new(size_t i_nSize, const char* i_pcFile, int i_nLine)
    {
      return HeapMgr::instance().alloc(i_nSize, i_pcFile, i_nLine);
    }

    void operator delete(void* i_paObject)
    {
      HeapMgr::instance().free(i_paObject);
    }

};

#define new new (__FILE__,__LINE__) 

class A : public HeapControl
{
  int m_nInt;
};

class B : public A
{
  float m_nFloat;
};

class C : public HeapControl
{
  double m_nDouble;
};

void test()
{
  A* paA = new A();
  A* paB = new B();
  C* paC = new C();

  delete paC;

  HeapMgr::instance().printStatus();
}

#undef new


}

// ----------------------------------------------------------------------------

main()
{
  cout << "--------------------------- Test 1 ---------------------------" << endl;
  LeakHunter1::test();

  cout << "--------------------------- Test 2 ---------------------------" << endl;
  LeakHunter2::test();

  cout << "--------------------------- Test 3 ---------------------------" << endl;
  LeakHunter3::test();

  cout << "--------------------------- Test 4 ---------------------------" << endl;
  LeakHunter4::test();

  cout << "--------------------------- Test 5 ---------------------------" << endl;
  LeakHunter5::test();

  return 0;
}